@@ -34,6 +34,78 @@
 #include <string.h>
 #include <sys/stat.h>
 
+/** Caching bfs_stat(). */
+static const struct bfs_stat *bftw_stat_impl(struct BFTW *ftwbuf, struct bftw_stat *cache, enum bfs_stat_flags flags) {
+	if (!cache->buf) {
+		if (cache->error) {
+			errno = cache->error;
+		} else if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, flags, &cache->storage) == 0) {
+			cache->buf = &cache->storage;
+		} else {
+			cache->error = errno;
+		}
+	}
+
+	return cache->buf;
+}
+
+const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
+	struct BFTW *mutbuf = (struct BFTW *)ftwbuf;
+	const struct bfs_stat *ret;
+
+	if (flags & BFS_STAT_NOFOLLOW) {
+		ret = bftw_stat_impl(mutbuf, &mutbuf->lstat_cache, BFS_STAT_NOFOLLOW);
+		if (ret && !S_ISLNK(ret->mode) && !mutbuf->stat_cache.buf) {
+			// Non-link, so share stat info
+			mutbuf->stat_cache.buf = ret;
+		}
+	} else {
+		ret = bftw_stat_impl(mutbuf, &mutbuf->stat_cache, BFS_STAT_FOLLOW);
+		if (!ret && (flags & BFS_STAT_TRYFOLLOW) && is_nonexistence_error(errno)) {
+			ret = bftw_stat_impl(mutbuf, &mutbuf->lstat_cache, BFS_STAT_NOFOLLOW);
+		}
+	}
+
+	return ret;
+}
+
+const struct bfs_stat *bftw_cached_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
+	if (flags & BFS_STAT_NOFOLLOW) {
+		return ftwbuf->lstat_cache.buf;
+	} else if (ftwbuf->stat_cache.buf) {
+		return ftwbuf->stat_cache.buf;
+	} else if ((flags & BFS_STAT_TRYFOLLOW) && is_nonexistence_error(ftwbuf->stat_cache.error)) {
+		return ftwbuf->lstat_cache.buf;
+	} else {
+		return NULL;
+	}
+}
+
+enum bfs_type bftw_type(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
+	if (flags & BFS_STAT_NOFOLLOW) {
+		if (ftwbuf->type == BFS_LNK || (ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
+			return ftwbuf->type;
+		}
+	} else if (flags & BFS_STAT_TRYFOLLOW) {
+		if (ftwbuf->type != BFS_LNK || (ftwbuf->stat_flags & BFS_STAT_TRYFOLLOW)) {
+			return ftwbuf->type;
+		}
+	} else {
+		if (ftwbuf->type != BFS_LNK) {
+			return ftwbuf->type;
+		} else if (ftwbuf->stat_flags & BFS_STAT_TRYFOLLOW) {
+			return BFS_ERROR;
+		}
+	}
+
+	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, flags);
+	if (statbuf) {
+		return bfs_mode_to_type(statbuf->mode);
+	} else {
+		return BFS_ERROR;
+	}
+}
+
 /**
  * A file.
  */
@@ -42,10 +114,13 @@ struct bftw_file {
 	struct bftw_file *parent;
 	/** The root under which this file was found. */
 	struct bftw_file *root;
-	/** The next file in the queue, if any. */
-	struct bftw_file *next;
 
-	/** LRU list links. */
+	/** The next file to open/close/visit. */
+	struct bftw_file *next;
+	/** The next directory to read. */
+	struct { struct bftw_file *next; } to_read;
+
+	/** LRU list node. */
 	struct {
 		struct bftw_file *prev;
 		struct bftw_file *next;
@@ -116,6 +191,16 @@ static void bftw_cache_init(struct bftw_cache *cache, size_t capacity) {
 	bfs_dir_arena(&cache->dirs);
 }
 
+/** Allocate a directory. */
+static struct bfs_dir *bftw_allocdir(struct bftw_cache *cache) {
+	return arena_alloc(&cache->dirs);
+}
+
+/** Free a directory. */
+static void bftw_freedir(struct bftw_cache *cache, struct bfs_dir *dir) {
+	arena_free(&cache->dirs, dir);
+}
+
 /** Remove a bftw_file from the LRU list. */
 static void bftw_lru_remove(struct bftw_cache *cache, struct bftw_file *file) {
 	if (cache->target == file) {
@@ -137,9 +222,8 @@ static void bftw_file_close(struct bftw_cache *cache, struct bftw_file *file) {
 	bfs_assert(file->pincount == 0);
 
 	if (file->dir) {
-		bfs_assert(file->fd == bfs_dirfd(file->dir));
 		bfs_closedir(file->dir);
-		arena_free(&cache->dirs, file->dir);
+		bftw_freedir(cache, file->dir);
 		file->dir = NULL;
 	} else {
 		xclose(file->fd);
@@ -249,6 +333,7 @@ static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_fil
 	}
 
 	file->next = NULL;
+	file->to_read.next = NULL;
 	file->lru.prev = file->lru.next = NULL;
 
 	file->refcount = 1;
@@ -267,143 +352,19 @@ static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_fil
 	return file;
 }
 
-/**
- * Open a bftw_file relative to another one.
- *
- * @param cache
- *         The cache to hold the file.
- * @param file
- *         The file to open.
- * @param base
- *         The base directory for the relative path (may be NULL).
- * @param at_fd
- *         The base file descriptor, AT_FDCWD if base == NULL.
- * @param at_path
- *         The relative path to the file.
- * @return
- *         The opened file descriptor, or negative on error.
- */
-static int bftw_file_openat(struct bftw_cache *cache, struct bftw_file *file, struct bftw_file *base, const char *at_path) {
-	bfs_assert(file->fd < 0);
-
-	int at_fd = AT_FDCWD;
-	if (base) {
-		bftw_cache_pin(cache, base);
-		at_fd = base->fd;
-	}
-
-	int flags = O_RDONLY | O_CLOEXEC | O_DIRECTORY;
-	int fd = openat(at_fd, at_path, flags);
-
-	if (fd < 0 && errno == EMFILE) {
-		if (bftw_cache_pop(cache) == 0) {
-			fd = openat(at_fd, at_path, flags);
-		}
-		cache->capacity = 1;
-	}
-
-	if (base) {
-		bftw_cache_unpin(cache, base);
-	}
-
-	if (fd >= 0) {
-		file->fd = fd;
-		bftw_cache_add(cache, file);
-	}
-
-	return file->fd;
-}
-
-/**
- * Open a bftw_file.
- *
- * @param cache
- *         The cache to hold the file.
- * @param file
- *         The file to open.
- * @param path
- *         The full path to the file.
- * @return
- *         The opened file descriptor, or negative on error.
- */
-static int bftw_file_open(struct bftw_cache *cache, struct bftw_file *file, const char *path) {
-	// Find the nearest open ancestor
-	struct bftw_file *base = file;
-	do {
-		base = base->parent;
-	} while (base && base->fd < 0);
-
-	const char *at_path = path;
-	if (base) {
-		at_path += bftw_child_nameoff(base);
-	}
-
-	int fd = bftw_file_openat(cache, file, base, at_path);
-	if (fd >= 0 || errno != ENAMETOOLONG) {
-		return fd;
-	}
-
-	// Handle ENAMETOOLONG by manually traversing the path component-by-component
-	struct bftw_list parents;
-	SLIST_INIT(&parents);
-
-	struct bftw_file *cur;
-	for (cur = file; cur != base; cur = cur->parent) {
-		SLIST_PREPEND(&parents, cur);
-	}
-
-	while ((cur = SLIST_POP(&parents))) {
-		if (!cur->parent || cur->parent->fd >= 0) {
-			bftw_file_openat(cache, cur, cur->parent, cur->name);
-		}
-	}
-
-	return file->fd;
-}
-
-/**
- * Associate an open directory with a bftw_file.
- */
+/** Associate an open directory with a bftw_file. */
 static void bftw_file_set_dir(struct bftw_cache *cache, struct bftw_file *file, struct bfs_dir *dir) {
 	bfs_assert(!file->dir);
 	file->dir = dir;
 
-	if (file->fd < 0) {
+	if (file->fd >= 0) {
+		bfs_assert(file->fd == bfs_dirfd(dir));
+	} else {
 		file->fd = bfs_dirfd(dir);
 		bftw_cache_add(cache, file);
 	}
-}
 
-/**
- * Open a bftw_file as a directory.
- *
- * @param cache
- *         The cache to hold the file.
- * @param file
- *         The directory to open.
- * @param path
- *         The full path to the directory.
- * @return
- *         The opened directory, or NULL on error.
- */
-static struct bfs_dir *bftw_file_opendir(struct bftw_cache *cache, struct bftw_file *file, const char *path) {
-	int fd = bftw_file_open(cache, file, path);
-	if (fd < 0) {
-		return NULL;
-	}
-
-	struct bfs_dir *dir = arena_alloc(&cache->dirs);
-	if (!dir) {
-		return NULL;
-	}
-
-	if (bfs_opendir(dir, fd, NULL) != 0) {
-		arena_free(&cache->dirs, dir);
-		return NULL;
-	}
-
-	bftw_file_set_dir(cache, file, dir);
-	return dir;
+	bftw_cache_pin(cache, file);
 }
 
 /** Free a bftw_file. */
@@ -443,10 +404,17 @@ struct bftw_state {
 	/** The number of I/O threads. */
 	size_t nthreads;
 
+	/** The queue of directories to open. */
+	struct bftw_list to_open;
 	/** The queue of directories to read. */
-	struct bftw_list dirs;
+	struct bftw_list to_read;
+	/** The queue of unpinned directories to unwrap. */
+	struct bftw_list to_close;
+	/** Available capacity in to_read. */
+	size_t dirlimit;
+
 	/** The queue of files to visit. */
-	struct bftw_list files;
+	struct bftw_list to_visit;
 	/** A batch of files to enqueue. */
 	struct bftw_list batch;
 
@@ -470,9 +438,7 @@ struct bftw_state {
 	struct BFTW ftwbuf;
 };
 
-/**
- * Initialize the bftw() state.
- */
+/** Initialize the bftw() state. */
 static int bftw_state_init(struct bftw_state *state, const struct bftw_args *args) {
 	state->callback = args->callback;
 	state->ptr = args->ptr;
@@ -490,38 +456,32 @@ static int bftw_state_init(struct bftw_state *state, const struct bftw_args *arg
 		errno = EMFILE;
 		return -1;
 	}
-
-	state->path = dstralloc(0);
-	if (!state->path) {
-		return -1;
-	}
-
 	bftw_cache_init(&state->cache, args->nopenfd);
 
-	size_t qdepth = args->nopenfd - 1;
-	if (qdepth > 1024) {
-		qdepth = 1024;
-	}
-
-	size_t nthreads = args->nthreads;
-	if (nthreads > qdepth) {
-		nthreads = qdepth;
-	}
-
-	state->ioq = NULL;
-	if (nthreads > 0) {
-		state->ioq = ioq_create(qdepth, nthreads);
+	state->nthreads = args->nthreads;
+	if (state->nthreads > 0) {
+		state->ioq = ioq_create(4096, state->nthreads);
 		if (!state->ioq) {
-			dstrfree(state->path);
 			return -1;
 		}
+	} else {
+		state->ioq = NULL;
 	}
-	state->nthreads = nthreads;
 
-	SLIST_INIT(&state->dirs);
-	SLIST_INIT(&state->files);
+	SLIST_INIT(&state->to_open);
+	SLIST_INIT(&state->to_read);
+	SLIST_INIT(&state->to_close);
+
+	size_t dirlimit = args->nopenfd - 1;
+	if (dirlimit > 1024) {
+		dirlimit = 1024;
+	}
+	state->dirlimit = dirlimit;
+
+	SLIST_INIT(&state->to_visit);
 	SLIST_INIT(&state->batch);
 
+	state->path = NULL;
 	state->file = NULL;
 	state->previous = NULL;
 
@@ -532,81 +492,401 @@ static int bftw_state_init(struct bftw_state *state, const struct bftw_args *arg
 	return 0;
 }
 
-/** Cached bfs_stat(). */
-static const struct bfs_stat *bftw_stat_impl(struct BFTW *ftwbuf, struct bftw_stat *cache, enum bfs_stat_flags flags) {
-	if (!cache->buf) {
-		if (cache->error) {
-			errno = cache->error;
-		} else if (bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, flags, &cache->storage) == 0) {
-			cache->buf = &cache->storage;
-		} else {
-			cache->error = errno;
-		}
+/** Pop a response from the I/O queue. */
+static int bftw_ioq_pop(struct bftw_state *state, bool block) {
+	struct ioq *ioq = state->ioq;
+	if (!ioq) {
+		return -1;
 	}
 
-	return cache->buf;
+	struct ioq_ent *ent = block ? ioq_pop(ioq) : ioq_trypop(ioq);
+	if (!ent) {
+		return -1;
+	}
+
+	struct bftw_cache *cache = &state->cache;
+	struct bftw_file *file;
+	struct bftw_file *parent;
+	struct bfs_dir *dir;
+
+	enum ioq_op op = ent->op;
+	switch (op) {
+	case IOQ_CLOSE:
+		++cache->capacity;
+		break;
+
+	case IOQ_CLOSEDIR:
+		++cache->capacity;
+		dir = ent->closedir.dir;
+		bftw_freedir(cache, dir);
+		break;
+
+	case IOQ_OPENDIR:
+		file = ent->ptr;
+		file->ioqueued = false;
+
+		++cache->capacity;
+		parent = file->parent;
+		if (parent) {
+			bftw_cache_unpin(cache, parent);
+			if (parent->pincount == 0 && parent->dir) {
+				SLIST_APPEND(&state->to_close, parent);
+			}
+		}
+
+		dir = ent->opendir.dir;
+		if (ent->ret == 0) {
+			bftw_file_set_dir(cache, file, dir);
+		} else {
+			bftw_freedir(cache, dir);
+			++state->dirlimit;
+		}
+
+		if (!(state->flags & BFTW_SORT)) {
+			SLIST_APPEND(&state->to_read, file, to_read);
+		}
+		break;
+	}
+
+	ioq_free(ioq, ent);
+	return op;
 }
 
-const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
-	struct BFTW *mutbuf = (struct BFTW *)ftwbuf;
-	const struct bfs_stat *ret;
+/** Try to reserve space in the I/O queue. */
+static int bftw_ioq_reserve(struct bftw_state *state) {
+	struct ioq *ioq = state->ioq;
+	if (!ioq) {
+		return -1;
+	}
 
-	if (flags & BFS_STAT_NOFOLLOW) {
-		ret = bftw_stat_impl(mutbuf, &mutbuf->lstat_cache, BFS_STAT_NOFOLLOW);
-		if (ret && !S_ISLNK(ret->mode) && !mutbuf->stat_cache.buf) {
-			// Non-link, so share stat info
-			mutbuf->stat_cache.buf = ret;
-		}
-	} else {
-		ret = bftw_stat_impl(mutbuf, &mutbuf->stat_cache, BFS_STAT_FOLLOW);
-		if (!ret && (flags & BFS_STAT_TRYFOLLOW) && is_nonexistence_error(errno)) {
-			ret = bftw_stat_impl(mutbuf, &mutbuf->lstat_cache, BFS_STAT_NOFOLLOW);
+	if (ioq_capacity(ioq) > 0) {
+		return 0;
+	}
+
+	// With more than two threads, it is faster to wait for an I/O operation
+	// to complete than it is to do it ourselves
+	bool block = state->nthreads > 2;
+	if (bftw_ioq_pop(state, block) < 0) {
+		return -1;
+	}
+
+	return 0;
+}
+
+/** Try to reserve space in the cache. */
+static int bftw_cache_reserve(struct bftw_state *state) {
+	struct bftw_cache *cache = &state->cache;
+	if (cache->capacity > 0) {
+		return 0;
+	}
+
+	while (bftw_ioq_pop(state, true) >= 0) {
+		if (cache->capacity > 0) {
+			return 0;
 		}
 	}
 
+	if (bftw_cache_pop(cache) != 0) {
+		errno = EMFILE;
+		return -1;
+	}
+
+	bfs_assert(cache->capacity > 0);
+	return 0;
+}
+
+/** Open a bftw_file relative to another one. */
+static int bftw_file_openat(struct bftw_state *state, struct bftw_file *file, struct bftw_file *base, const char *at_path) {
+	bfs_assert(file->fd < 0);
+
+	struct bftw_cache *cache = &state->cache;
+
+	int at_fd = AT_FDCWD;
+	if (base) {
+		bftw_cache_pin(cache, base);
+		at_fd = base->fd;
+	}
+
+	int fd = -1;
+	if (bftw_cache_reserve(state) != 0) {
+		goto unpin;
+	}
+
+	int flags = O_RDONLY | O_CLOEXEC | O_DIRECTORY;
+	fd = openat(at_fd, at_path, flags);
+
+	if (fd < 0 && errno == EMFILE) {
+		if (bftw_cache_pop(cache) == 0) {
+			fd = openat(at_fd, at_path, flags);
+		}
+		cache->capacity = 1;
+	}
+
+unpin:
+	if (base) {
+		bftw_cache_unpin(cache, base);
+	}
+
+	if (fd >= 0) {
+		file->fd = fd;
+		bftw_cache_add(cache, file);
+	}
+
+	return fd;
+}
+
+/** Open a bftw_file. */
+static int bftw_file_open(struct bftw_state *state, struct bftw_file *file, const char *path) {
+	// Find the nearest open ancestor
+	struct bftw_file *base = file;
+	do {
+		base = base->parent;
+	} while (base && base->fd < 0);
+
+	const char *at_path = path;
+	if (base) {
+		at_path += bftw_child_nameoff(base);
+	}
+
+	int fd = bftw_file_openat(state, file, base, at_path);
+	if (fd >= 0 || errno != ENAMETOOLONG) {
+		return fd;
+	}
+
+	// Handle ENAMETOOLONG by manually traversing the path component-by-component
+	struct bftw_list parents;
+	SLIST_INIT(&parents);
+
+	struct bftw_file *cur;
+	for (cur = file; cur != base; cur = cur->parent) {
+		SLIST_PREPEND(&parents, cur);
+	}
+
+	while ((cur = SLIST_POP(&parents))) {
+		if (!cur->parent || cur->parent->fd >= 0) {
+			bftw_file_openat(state, cur, cur->parent, cur->name);
+		}
+	}
+
+	return file->fd;
+}
+
+/** Close a directory, asynchronously if possible. */
+static int bftw_ioq_closedir(struct bftw_state *state, struct bfs_dir *dir) {
+	if (bftw_ioq_reserve(state) == 0) {
+		if (ioq_closedir(state->ioq, dir, NULL) == 0) {
+			return 0;
+		}
+	}
+
+	struct bftw_cache *cache = &state->cache;
+	int ret = bfs_closedir(dir);
+	bftw_freedir(cache, dir);
+	++cache->capacity;
 	return ret;
 }
 
-const struct bfs_stat *bftw_cached_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
-	if (flags & BFS_STAT_NOFOLLOW) {
-		return ftwbuf->lstat_cache.buf;
-	} else if (ftwbuf->stat_cache.buf) {
-		return ftwbuf->stat_cache.buf;
-	} else if ((flags & BFS_STAT_TRYFOLLOW) && is_nonexistence_error(ftwbuf->stat_cache.error)) {
-		return ftwbuf->lstat_cache.buf;
+/** Close a file descriptor, asynchronously if possible. */
+static int bftw_ioq_close(struct bftw_state *state, int fd) {
+	if (bftw_ioq_reserve(state) == 0) {
+		if (ioq_close(state->ioq, fd, NULL) == 0) {
+			return 0;
+		}
+	}
+
+	struct bftw_cache *cache = &state->cache;
+	int ret = xclose(fd);
+	++cache->capacity;
+	return ret;
+}
+
+/** Close a file, asynchronously if possible. */
+static int bftw_close(struct bftw_state *state, struct bftw_file *file) {
+	bfs_assert(file->fd >= 0);
+	bfs_assert(file->pincount == 0);
+
+	struct bfs_dir *dir = file->dir;
+	int fd = file->fd;
+
+	bftw_lru_remove(&state->cache, file);
+	file->dir = NULL;
+	file->fd = -1;
+
+	if (dir) {
+		return bftw_ioq_closedir(state, dir);
 	} else {
-		return NULL;
+		return bftw_ioq_close(state, fd);
 	}
 }
 
-enum bfs_type bftw_type(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
-	if (flags & BFS_STAT_NOFOLLOW) {
-		if (ftwbuf->type == BFS_LNK || (ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
-			return ftwbuf->type;
-		}
-	} else if (flags & BFS_STAT_TRYFOLLOW) {
-		if (ftwbuf->type != BFS_LNK || (ftwbuf->stat_flags & BFS_STAT_TRYFOLLOW)) {
-			return ftwbuf->type;
-		}
-	} else {
-		if (ftwbuf->type != BFS_LNK) {
-			return ftwbuf->type;
-		} else if (ftwbuf->stat_flags & BFS_STAT_TRYFOLLOW) {
-			return BFS_ERROR;
-		}
+/** Free an open directory. */
+static int bftw_unwrapdir(struct bftw_state *state, struct bftw_file *file) {
+	struct bfs_dir *dir = file->dir;
+	if (!dir) {
+		return 0;
 	}
 
-	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, flags);
-	if (statbuf) {
-		return bfs_mode_to_type(statbuf->mode);
-	} else {
-		return BFS_ERROR;
+	struct bftw_cache *cache = &state->cache;
+
+	// Try to keep an open fd if any children exist
+	bool reffed = file->refcount > 1;
+	// Keep the fd the same if it's pinned
+	bool pinned = file->pincount > 0;
+
+#if BFS_USE_UNWRAPDIR
+	if (reffed || pinned) {
+		bfs_unwrapdir(dir);
+		bftw_freedir(cache, dir);
+		file->dir = NULL;
+		return 0;
+	}
+#else
+	if (pinned) {
+		return -1;
+	}
+#endif
+
+	if (!reffed) {
+		return bftw_close(state, file);
+	}
+
+	if (bftw_cache_reserve(state) != 0) {
+		return -1;
+	}
+
+	int fd = dup_cloexec(file->fd);
+	if (fd < 0) {
+		return -1;
+	}
+	--cache->capacity;
+
+	file->dir = NULL;
+	file->fd = fd;
+	return bftw_ioq_closedir(state, dir);
+}
+
+/** Open a directory asynchronously. */
+static int bftw_ioq_opendir(struct bftw_state *state, struct bftw_file *file) {
+	if (state->dirlimit == 0) {
+		goto fail;
+	}
+
+	if (bftw_ioq_reserve(state) != 0) {
+		goto fail;
+	}
+
+	int dfd = AT_FDCWD;
+	struct bftw_cache *cache = &state->cache;
+	struct bftw_file *parent = file->parent;
+	if (parent) {
+		dfd = parent->fd;
+		if (dfd < 0) {
+			goto fail;
+		}
+		bftw_cache_pin(cache, parent);
+	}
+
+	if (bftw_cache_reserve(state) != 0) {
+		goto unpin;
+	}
+
+	struct bfs_dir *dir = bftw_allocdir(cache);
+	if (!dir) {
+		goto unpin;
+	}
+
+	if (ioq_opendir(state->ioq, dir, dfd, file->name, file) != 0) {
+		goto free;
+	}
+
+	file->ioqueued = true;
+	--cache->capacity;
+	--state->dirlimit;
+	return 0;
+
+free:
+	bftw_freedir(cache, dir);
+unpin:
+	if (parent) {
+		bftw_cache_unpin(cache, parent);
+	}
+fail:
+	return -1;
+}
+
+/** Push a directory onto the queue. */
+static void bftw_push_dir(struct bftw_state *state, struct bftw_file *file) {
+	bfs_assert(file->type == BFS_DIR);
+
+	SLIST_APPEND(&state->to_open, file);
+
+	if (state->flags & BFTW_SORT) {
+		// When sorting, directories are kept in order on the to_read
+		// list; otherwise, they are only added once they are open
+		SLIST_APPEND(&state->to_read, file, to_read);
+	}
+
+	while (state->to_open.head) {
+		if (bftw_ioq_opendir(state, state->to_open.head) == 0) {
+			SLIST_POP(&state->to_open);
+		} else {
+			break;
+		}
 	}
 }
 
-/**
- * Build the path to the current file.
- */
+/** Pop a directory to read from the queue. */
+static bool bftw_pop_dir(struct bftw_state *state) {
+	bfs_assert(!state->file);
+
+	bool have_files = state->to_visit.head;
+
+	if (state->flags & BFTW_SORT) {
+		// Keep strict breadth-first order when sorting
+		if (state->strategy != BFTW_DFS && have_files) {
+			return false;
+		}
+	} else {
+		while (!state->to_read.head) {
+			// Block if we have no other files/dirs to visit, or no room in the cache
+			bool have_dirs = state->to_open.head;
+			bool have_room = state->cache.capacity > 0;
+			bool block = !(have_dirs || have_files) || !have_room;
+
+			if (bftw_ioq_pop(state, block) < 0) {
+				break;
+			}
+		}
+	}
+
+	struct bftw_file *file = SLIST_POP(&state->to_read, to_read);
+	if (!file || file == state->to_open.head) {
+		file = SLIST_POP(&state->to_open);
+	}
+	if (!file) {
+		return false;
+	}
+
+	if (file->dir) {
+		++state->dirlimit;
+	}
+
+	while (file->ioqueued) {
+		bftw_ioq_pop(state, true);
+	}
+
+	state->file = file;
+	return true;
+}
+
+/** Pop a file to visit from the queue. */
+static bool bftw_pop_file(struct bftw_state *state) {
+	bfs_assert(!state->file);
+	state->file = SLIST_POP(&state->to_visit);
+	return state->file;
+}
+
+/** Build the path to the current file. */
 static int bftw_build_path(struct bftw_state *state, const char *name) {
 	const struct bftw_file *file = state->file;
 
@@ -653,6 +933,72 @@ static int bftw_build_path(struct bftw_state *state, const char *name) {
 	return 0;
 }
 
+/** Open a bftw_file as a directory. */
+static struct bfs_dir *bftw_file_opendir(struct bftw_state *state, struct bftw_file *file, const char *path) {
+	int fd = bftw_file_open(state, file, path);
+	if (fd < 0) {
+		return NULL;
+	}
+
+	struct bftw_cache *cache = &state->cache;
+	struct bfs_dir *dir = bftw_allocdir(cache);
+	if (!dir) {
+		return NULL;
+	}
+
+	if (bfs_opendir(dir, fd, NULL) != 0) {
+		bftw_freedir(cache, dir);
+		return NULL;
+	}
+
+	bftw_file_set_dir(cache, file, dir);
+	return dir;
+}
+
+/** Open the current directory. */
+static int bftw_opendir(struct bftw_state *state) {
+	bfs_assert(!state->dir);
+	bfs_assert(!state->de);
+
+	state->direrror = 0;
+
+	struct bftw_file *file = state->file;
+	state->dir = file->dir;
+	if (state->dir) {
+		return 0;
+	}
+
+	if (bftw_build_path(state, NULL) != 0) {
+		return -1;
+	}
+
+	state->dir = bftw_file_opendir(state, file, state->path);
+	if (!state->dir) {
+		state->direrror = errno;
+	}
+
+	return 0;
+}
+
+/** Read an entry from the current directory. */
+static int bftw_readdir(struct bftw_state *state) {
+	if (!state->dir) {
+		return -1;
+	}
+
+	int ret = bfs_readdir(state->dir, &state->de_storage);
+	if (ret > 0) {
+		state->de = &state->de_storage;
+	} else if (ret == 0) {
+		state->de = NULL;
+	} else {
+		state->de = NULL;
+		state->direrror = errno;
+	}
+
+	return ret;
+}
+
 /** Check if a stat() call is needed for this visit. */
 static bool bftw_need_stat(const struct bftw_state *state) {
 	if (state->flags & BFTW_STAT) {
@@ -694,17 +1040,8 @@ static void bftw_stat_init(struct bftw_stat *cache) {
 	cache->error = 0;
 }
 
-/**
- * Open a file if necessary.
- *
- * @param file
- *         The file to open.
- * @param path
- *         The path to that file or one of its descendants.
- * @return
- *         The opened file descriptor, or -1 on error.
- */
-static int bftw_ensure_open(struct bftw_cache *cache, struct bftw_file *file, const char *path) {
+/** Open a file if necessary. */
+static int bftw_ensure_open(struct bftw_state *state, struct bftw_file *file, const char *path) {
 	int ret = file->fd;
 
 	if (ret < 0) {
@@ -713,16 +1050,14 @@ static int bftw_ensure_open(struct bftw_cache *cache, struct bftw_file *file, co
 			return -1;
 		}
 
-		ret = bftw_file_open(cache, file, copy);
+		ret = bftw_file_open(state, file, copy);
 		free(copy);
 	}
 
 	return ret;
 }
 
-/**
- * Initialize the buffers with data about the current path.
- */
+/** Initialize the buffers with data about the current path. */
 static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	struct bftw_file *file = state->file;
 	const struct bfs_dirent *de = state->de;
@@ -755,7 +1090,7 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 
 	if (parent) {
 		// Try to ensure the immediate parent is open, to avoid ENAMETOOLONG
-		if (bftw_ensure_open(&state->cache, parent, state->path) >= 0) {
+		if (bftw_ensure_open(state, parent, state->path) >= 0) {
 			ftwbuf->at_fd = parent->fd;
 			ftwbuf->at_path += ftwbuf->nameoff;
 		} else {
@@ -822,9 +1157,7 @@ static bool bftw_is_mount(struct bftw_state *state, const char *name) {
 	return statbuf && statbuf->dev != parent->dev;
 }
 
-/**
- * Invoke the callback.
- */
+/** Invoke the callback. */
 static enum bftw_action bftw_call_back(struct bftw_state *state, const char *name, enum bftw_visit visit) {
 	if (visit == BFTW_POST && !(state->flags & BFTW_POST_ORDER)) {
 		return BFTW_PRUNE;
@@ -870,340 +1203,6 @@ static enum bftw_action bftw_call_back(struct bftw_state *state, const char *nam
 	}
 }
 
-/** Pop a response from the I/O queue. */
-static int bftw_ioq_pop(struct bftw_state *state, bool block) {
-	struct ioq *ioq = state->ioq;
-	if (!ioq) {
-		return -1;
-	}
-
-	struct ioq_ent *ent = block ? ioq_pop(ioq) : ioq_trypop(ioq);
-	if (!ent) {
-		return -1;
-	}
-
-	struct bftw_cache *cache = &state->cache;
-	struct bftw_file *file;
-	struct bfs_dir *dir;
-
-	enum ioq_op op = ent->op;
-	switch (op) {
-	case IOQ_CLOSE:
-		++cache->capacity;
-		break;
-
-	case IOQ_CLOSEDIR:
-		++cache->capacity;
-		dir = ent->closedir.dir;
-		arena_free(&cache->dirs, dir);
-		break;
-
-	case IOQ_OPENDIR:
-		file = ent->ptr;
-		file->ioqueued = false;
-
-		++cache->capacity;
-		if (file->parent) {
-			bftw_cache_unpin(cache, file->parent);
-		}
-
-		dir = ent->opendir.dir;
-		if (ent->ret == 0) {
-			bftw_file_set_dir(cache, file, dir);
-		} else {
-			arena_free(&cache->dirs, dir);
-		}
-
-		if (!(state->flags & BFTW_SORT)) {
-			SLIST_PREPEND(&state->dirs, file);
-		}
-		break;
-	}
-
-	ioq_free(ioq, ent);
-	return op;
-}
-
-/** Try to reserve space in the I/O queue. */
-static int bftw_ioq_reserve(struct bftw_state *state) {
-	struct ioq *ioq = state->ioq;
-	if (!ioq) {
-		return -1;
-	}
-
-	if (ioq_capacity(ioq) > 0) {
-		return 0;
-	}
-
-	// With more than two threads, it is faster to wait for an I/O operation
-	// to complete than it is to do it ourselves
-	bool block = state->nthreads > 2;
-	if (bftw_ioq_pop(state, block) < 0) {
-		return -1;
-	}
-
-	return 0;
-}
-
-/** Try to reserve space in the cache. */
-static int bftw_cache_reserve(struct bftw_state *state) {
-	struct bftw_cache *cache = &state->cache;
-	if (cache->capacity > 0) {
-		return 0;
-	}
-
-	while (bftw_ioq_pop(state, false) >= 0) {
-		if (cache->capacity > 0) {
-			return 0;
-		}
-	}
-
-	return bftw_cache_pop(cache);
-}
-
-/** Push a directory onto the queue. */
-static void bftw_push_dir(struct bftw_state *state, struct bftw_file *file) {
-	bfs_assert(file->type == BFS_DIR);
-
-	struct bftw_cache *cache = &state->cache;
-
-	if (bftw_ioq_reserve(state) != 0) {
-		goto append;
-	}
-
-	int dfd = AT_FDCWD;
-	if (file->parent) {
-		dfd = file->parent->fd;
-		if (dfd < 0) {
-			goto append;
-		}
-		bftw_cache_pin(cache, file->parent);
-	}
-
-	if (bftw_cache_reserve(state) != 0) {
-		goto unpin;
-	}
-
-	struct bfs_dir *dir = arena_alloc(&cache->dirs);
-	if (!dir) {
-		goto unpin;
-	}
-
-	if (ioq_opendir(state->ioq, dir, dfd, file->name, file) != 0) {
-		goto free;
-	}
-
-	file->ioqueued = true;
-	--cache->capacity;
-
-	if (state->flags & BFTW_SORT) {
-		// When sorting, dirs are always kept in order in state->dirs,
-		// and we wait for the ioq to open them before popping
-		goto append;
-	} else {
-		// When not sorting, dirs are not added to state->dirs until
-		// popped from the ioq
-		return;
-	}
-
-free:
-	arena_free(&cache->dirs, dir);
-unpin:
-	if (file->parent) {
-		bftw_cache_unpin(cache, file->parent);
-	}
-append:
-	SLIST_APPEND(&state->dirs, file);
-}
-
-/** Close a directory, asynchronously if possible. */
-static int bftw_ioq_closedir(struct bftw_state *state, struct bfs_dir *dir) {
-	if (bftw_ioq_reserve(state) == 0) {
-		if (ioq_closedir(state->ioq, dir, NULL) == 0) {
-			return 0;
-		}
-	}
-
-	struct bftw_cache *cache = &state->cache;
-	int ret = bfs_closedir(dir);
-	arena_free(&cache->dirs, dir);
-	++cache->capacity;
-	return ret;
-}
-
-/** Close a file descriptor, asynchronously if possible. */
-static int bftw_ioq_close(struct bftw_state *state, int fd) {
-	if (bftw_ioq_reserve(state) == 0) {
-		if (ioq_close(state->ioq, fd, NULL) == 0) {
-			return 0;
-		}
-	}
-
-	struct bftw_cache *cache = &state->cache;
-	int ret = xclose(fd);
-	++cache->capacity;
-	return ret;
-}
-
-/** Close a file, asynchronously if possible. */
-static int bftw_close(struct bftw_state *state, struct bftw_file *file) {
-	bfs_assert(file->fd >= 0);
-	bfs_assert(file->pincount == 0);
-
-	struct bfs_dir *dir = file->dir;
-	int fd = file->fd;
-
-	bftw_lru_remove(&state->cache, file);
-	file->dir = NULL;
-	file->fd = -1;
-
-	if (dir) {
-		return bftw_ioq_closedir(state, dir);
-	} else {
-		return bftw_ioq_close(state, fd);
-	}
-}
-
-/** Free an open directory. */
-static int bftw_unwrapdir(struct bftw_state *state, struct bftw_file *file) {
-	struct bfs_dir *dir = file->dir;
-	if (!dir) {
-		return 0;
-	}
-
-	struct bftw_cache *cache = &state->cache;
-
-	// Try to keep an open fd if any children exist
-	bool reffed = file->refcount > 1;
-	// Keep the fd the same if it's pinned
-	bool pinned = file->pincount > 0;
-
-#if BFS_USE_UNWRAPDIR
-	if (reffed || pinned) {
-		bfs_unwrapdir(dir);
-		arena_free(&cache->dirs, dir);
-		file->dir = NULL;
-		return 0;
-	}
-#else
-	if (pinned) {
-		return -1;
-	}
-#endif
-
-	if (!reffed) {
-		return bftw_close(state, file);
-	}
-
-	if (bftw_cache_reserve(state) != 0) {
-		return -1;
-	}
-
-	int fd = dup_cloexec(file->fd);
-	if (fd < 0) {
-		return -1;
-	}
-	--cache->capacity;
-
-	file->dir = NULL;
-	file->fd = fd;
-	return bftw_ioq_closedir(state, dir);
-}
-
-/** Pop a directory to read from the queue. */
-static bool bftw_pop_dir(struct bftw_state *state) {
-	bfs_assert(!state->file);
-
-	struct bftw_file *dir = state->dirs.head;
-	bool have_files = state->files.head;
-
-	if (state->flags & BFTW_SORT) {
-		// Keep strict breadth-first order when sorting
-		if (state->strategy != BFTW_DFS && have_files) {
-			return false;
-		}
-	} else {
-		while (!dir || !dir->dir) {
-			// Block if we have no other files/dirs to visit, or no room in the cache
-			bool have_room = state->cache.capacity > 0;
-			bool block = !(dir || have_files) || !have_room;
-
-			if (bftw_ioq_pop(state, block) < 0) {
-				break;
-			}
-			dir = state->dirs.head;
-		}
-	}
-
-	if (!dir) {
-		return false;
-	}
-
-	SLIST_POP(&state->dirs);
-	state->file = dir;
-
-	while (dir->ioqueued) {
-		bftw_ioq_pop(state, true);
-	}
-
-	return true;
-}
-
-/** Pop a file to visit from the queue. */
-static bool bftw_pop_file(struct bftw_state *state) {
-	bfs_assert(!state->file);
-	return (state->file = SLIST_POP(&state->files));
-}
-
-/**
- * Open the current directory.
- */
-static int bftw_opendir(struct bftw_state *state) {
-	bfs_assert(!state->dir);
-	bfs_assert(!state->de);
-
-	state->direrror = 0;
-
-	struct bftw_file *file = state->file;
-	if (file->dir) {
-		state->dir = file->dir;
-	} else {
-		if (bftw_build_path(state, NULL) != 0) {
-			return -1;
-		}
-		state->dir = bftw_file_opendir(&state->cache, file, state->path);
-	}
-
-	if (state->dir) {
-		bftw_cache_pin(&state->cache, file);
-	} else {
-		state->direrror = errno;
-	}
-
-	return 0;
-}
-
-/**
- * Read an entry from the current directory.
- */
-static int bftw_readdir(struct bftw_state *state) {
-	if (!state->dir) {
-		return -1;
-	}
-
-	int ret = bfs_readdir(state->dir, &state->de_storage);
-	if (ret > 0) {
-		state->de = &state->de_storage;
-	} else if (ret == 0) {
-		state->de = NULL;
-	} else {
-		state->de = NULL;
-		state->direrror = errno;
-	}
-
-	return ret;
-}
-
 /**
  * Flags controlling which files get visited when done with a directory.
  */
@@ -1220,15 +1219,14 @@ enum bftw_gc_flags {
 	BFTW_VISIT_ALL = BFTW_VISIT_ERROR | BFTW_VISIT_FILE | BFTW_VISIT_PARENTS,
 };
 
-/**
- * Garbage collect the current file and its parents.
- */
+/** Garbage collect the current file and its parents. */
 static int bftw_gc(struct bftw_state *state, enum bftw_gc_flags flags) {
 	int ret = 0;
 
-	if (state->dir) {
-		bftw_cache_unpin(&state->cache, state->file);
-		bftw_unwrapdir(state, state->file);
+	struct bftw_file *file = state->file;
+	if (file && file->dir) {
+		bftw_cache_unpin(&state->cache, file);
+		SLIST_APPEND(&state->to_close, file);
 	}
 	state->dir = NULL;
 	state->de = NULL;
@@ -1245,9 +1243,12 @@ static int bftw_gc(struct bftw_state *state, enum bftw_gc_flags flags) {
 	}
 	state->direrror = 0;
 
+	while ((file = SLIST_POP(&state->to_close))) {
+		bftw_unwrapdir(state, file);
+	}
+
 	enum bftw_gc_flags visit = BFTW_VISIT_FILE;
-	while (state->file) {
-		struct bftw_file *file = state->file;
+	while ((file = state->file)) {
 		if (--file->refcount > 0) {
 			state->file = NULL;
 			break;
@@ -1276,35 +1277,6 @@ static int bftw_gc(struct bftw_state *state, enum bftw_gc_flags flags) {
 	return ret;
 }
 
-/**
- * Dispose of the bftw() state.
- *
- * @return
- *         The bftw() return value.
- */
-static int bftw_state_destroy(struct bftw_state *state) {
-	dstrfree(state->path);
-
-	struct ioq *ioq = state->ioq;
-	if (ioq) {
-		ioq_cancel(ioq);
-		while (bftw_ioq_pop(state, true) >= 0);
-		state->ioq = NULL;
-	}
-
-	SLIST_EXTEND(&state->files, &state->batch);
-	do {
-		bftw_gc(state, BFTW_VISIT_NONE);
-	} while (bftw_pop_dir(state) || bftw_pop_file(state));
-
-	ioq_destroy(ioq);
-
-	bftw_cache_destroy(&state->cache);
-
-	errno = state->error;
-	return state->error ? -1 : 0;
-}
-
 /** Sort a bftw_list by filename. */
 static void bftw_list_sort(struct bftw_list *list) {
 	if (!list->head || !list->head->next) {
@@ -1350,9 +1322,9 @@ static void bftw_batch_finish(struct bftw_state *state) {
 	}
 
 	if (state->strategy != BFTW_BFS) {
-		SLIST_EXTEND(&state->batch, &state->files);
+		SLIST_EXTEND(&state->batch, &state->to_visit);
 	}
-	SLIST_EXTEND(&state->files, &state->batch);
+	SLIST_EXTEND(&state->to_visit, &state->batch);
 }
 
 /** Close the current directory. */
@@ -1426,6 +1398,35 @@ static int bftw_visit(struct bftw_state *state, const char *name) {
 	}
 }
 
+/**
+ * Dispose of the bftw() state.
+ *
+ * @return
+ *         The bftw() return value.
+ */
+static int bftw_state_destroy(struct bftw_state *state) {
+	dstrfree(state->path);
+
+	struct ioq *ioq = state->ioq;
+	if (ioq) {
+		ioq_cancel(ioq);
+		while (bftw_ioq_pop(state, true) >= 0);
+		state->ioq = NULL;
+	}
+
+	SLIST_EXTEND(&state->to_visit, &state->batch);
+	do {
+		bftw_gc(state, BFTW_VISIT_NONE);
+	} while (bftw_pop_dir(state) || bftw_pop_file(state));
+
+	ioq_destroy(ioq);
+
+	bftw_cache_destroy(&state->cache);
+
+	errno = state->error;
+	return state->error ? -1 : 0;
+}
+
 /**
  * bftw() implementation for simple breadth-/depth-first search.
  */
