@@ -35,11 +35,9 @@
 #include "exec.h"
 #include "expr.h"
 #include "pwcache.h"
-#include <assert.h>
 #include <errno.h>
 #include <limits.h>
 #include <stdarg.h>
-#include <stdbool.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -309,7 +307,7 @@ struct opt_state {
 /** Log an optimization. */
 BFS_FORMATTER(3, 4)
 static bool opt_debug(const struct opt_state *state, int level, const char *format, ...) {
-	assert(state->ctx->optlevel >= level);
+	bfs_assert(state->ctx->optlevel >= level);
 
 	if (bfs_debug(state->ctx, DEBUG_OPT, "${cyn}-O%d${rs}: ", level)) {
 		va_list args;
@@ -388,7 +386,7 @@ static struct bfs_expr *de_morgan(const struct opt_state *state, struct bfs_expr
 		has_parent = false;
 	}
 
-	assert(expr->eval_fn == eval_and || expr->eval_fn == eval_or);
+	bfs_assert(expr->eval_fn == eval_and || expr->eval_fn == eval_or);
 	if (expr->eval_fn == eval_and) {
 		expr->eval_fn = eval_or;
 		expr->argv = &fake_or_arg;
@@ -447,7 +445,7 @@ static struct bfs_expr *optimize_expr_recursive(struct opt_state *state, struct
  * Optimize a negation.
  */
 static struct bfs_expr *optimize_not_expr(const struct opt_state *state, struct bfs_expr *expr) {
-	assert(expr->eval_fn == eval_not);
+	bfs_assert(expr->eval_fn == eval_not);
 
 	struct bfs_expr *rhs = expr->rhs;
 
@@ -499,7 +497,7 @@ fail:
 
 /** Optimize a conjunction. */
 static struct bfs_expr *optimize_and_expr(const struct opt_state *state, struct bfs_expr *expr) {
-	assert(expr->eval_fn == eval_and);
+	bfs_assert(expr->eval_fn == eval_and);
 
 	struct bfs_expr *lhs = expr->lhs;
 	struct bfs_expr *rhs = expr->rhs;
@@ -570,7 +568,7 @@ fail:
 
 /** Optimize a disjunction. */
 static struct bfs_expr *optimize_or_expr(const struct opt_state *state, struct bfs_expr *expr) {
-	assert(expr->eval_fn == eval_or);
+	bfs_assert(expr->eval_fn == eval_or);
 
 	struct bfs_expr *lhs = expr->lhs;
 	struct bfs_expr *rhs = expr->rhs;
@@ -674,7 +672,7 @@ static struct bfs_expr *ignore_result(const struct opt_state *state, struct bfs_
 
 /** Optimize a comma expression. */
 static struct bfs_expr *optimize_comma_expr(const struct opt_state *state, struct bfs_expr *expr) {
-	assert(expr->eval_fn == eval_comma);
+	bfs_assert(expr->eval_fn == eval_comma);
 
 	struct bfs_expr *lhs = expr->lhs;
 	struct bfs_expr *rhs = expr->rhs;
@@ -1154,21 +1152,21 @@ static const struct {
  * Look up the appropriate optimizer for an expression and call it.
  */
 static struct bfs_expr *optimize_expr_lookup(struct opt_state *state, struct bfs_expr *expr) {
-	for (size_t i = 0; i < BFS_COUNTOF(opt_pure); ++i) {
+	for (size_t i = 0; i < countof(opt_pure); ++i) {
 		if (opt_pure[i] == expr->eval_fn) {
 			expr->pure = true;
 			break;
 		}
 	}
 
-	for (size_t i = 0; i < BFS_COUNTOF(opt_always_true); ++i) {
+	for (size_t i = 0; i < countof(opt_always_true); ++i) {
 		if (opt_always_true[i] == expr->eval_fn) {
 			expr->always_true = true;
 			break;
 		}
 	}
 
-	for (size_t i = 0; i < BFS_COUNTOF(opt_always_false); ++i) {
+	for (size_t i = 0; i < countof(opt_always_false); ++i) {
 		if (opt_always_false[i] == expr->eval_fn) {
 			expr->always_false = true;
 			break;
@@ -1176,35 +1174,35 @@ static struct bfs_expr *optimize_expr_lookup(struct opt_state *state, struct bfs
 	}
 
 	expr->cost = FAST_COST;
-	for (size_t i = 0; i < BFS_COUNTOF(opt_costs); ++i) {
+	for (size_t i = 0; i < countof(opt_costs); ++i) {
 		if (opt_costs[i].eval_fn == expr->eval_fn) {
 			expr->cost = opt_costs[i].cost;
 			break;
 		}
 	}
 
-	for (size_t i = 0; i < BFS_COUNTOF(opt_probs); ++i) {
+	for (size_t i = 0; i < countof(opt_probs); ++i) {
 		if (opt_probs[i].eval_fn == expr->eval_fn) {
 			expr->probability = opt_probs[i].probability;
 			break;
 		}
 	}
 
-	for (size_t i = 0; i < BFS_COUNTOF(opt_preds); ++i) {
+	for (size_t i = 0; i < countof(opt_preds); ++i) {
 		if (opt_preds[i].eval_fn == expr->eval_fn) {
 			infer_pred_facts(state, opt_preds[i].pred);
 			break;
 		}
 	}
 
-	for (size_t i = 0; i < BFS_COUNTOF(opt_ranges); ++i) {
+	for (size_t i = 0; i < countof(opt_ranges); ++i) {
 		if (opt_ranges[i].eval_fn == expr->eval_fn) {
 			infer_icmp_facts(state, expr, opt_ranges[i].range);
 			break;
 		}
 	}
 
-	for (size_t i = 0; i < BFS_COUNTOF(opt_fns); ++i) {
+	for (size_t i = 0; i < countof(opt_fns); ++i) {
 		if (opt_fns[i].eval_fn == expr->eval_fn) {
 			return opt_fns[i].opt_fn(state, expr);
 		}
@@ -1227,10 +1225,6 @@ static struct bfs_expr *optimize_expr_recursive(struct opt_state *state, struct
 		return ret;
 	}
 
-	if (!bfs_expr_is_parent(expr) && !expr->pure) {
-		facts_union(state->facts_when_impure, state->facts_when_impure, &state->facts);
-	}
-
 	expr = optimize_expr_lookup(state, expr);
 	if (!expr) {
 		return NULL;
@@ -1249,6 +1243,8 @@ static struct bfs_expr *optimize_expr_recursive(struct opt_state *state, struct
 				expr->ephemeral_fds = lhs->ephemeral_fds;
 			}
 		}
+	} else if (!expr->pure) {
+		facts_union(state->facts_when_impure, state->facts_when_impure, &state->facts);
 	}
 
 	if (expr->always_true) {
