@@ -9,6 +9,7 @@
  */
 
 #include "parse.h"
+#include "alloc.h"
 #include "bfstd.h"
 #include "bftw.h"
 #include "color.h"
@@ -24,12 +25,12 @@
 #include "opt.h"
 #include "printf.h"
 #include "pwcache.h"
+#include "sanity.h"
 #include "stat.h"
 #include "typo.h"
 #include "xregex.h"
 #include "xspawn.h"
 #include "xtime.h"
-#include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <fnmatch.h>
@@ -37,7 +38,6 @@
 #include <limits.h>
 #include <pwd.h>
 #include <stdarg.h>
-#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -56,39 +56,16 @@ static char *fake_print_arg = "-print";
 static char *fake_true_arg = "-true";
 
 struct bfs_expr *bfs_expr_new(bfs_eval_fn *eval_fn, size_t argc, char **argv) {
-	struct bfs_expr *expr = malloc(sizeof(*expr));
+	struct bfs_expr *expr = ZALLOC(struct bfs_expr);
 	if (!expr) {
-		perror("malloc()");
+		perror("zalloc()");
 		return NULL;
 	}
 
 	expr->eval_fn = eval_fn;
 	expr->argc = argc;
 	expr->argv = argv;
-	expr->persistent_fds = 0;
-	expr->ephemeral_fds = 0;
-	expr->pure = false;
-	expr->always_true = false;
-	expr->always_false = false;
-	expr->cost = 0.0;
 	expr->probability = 0.5;
-	expr->evaluations = 0;
-	expr->successes = 0;
-	expr->elapsed.tv_sec = 0;
-	expr->elapsed.tv_nsec = 0;
-
-	// Prevent bfs_expr_free() from freeing uninitialized pointers on error paths
-	if (bfs_expr_is_parent(expr)) {
-		expr->lhs = NULL;
-		expr->rhs = NULL;
-	} else if (eval_fn == eval_exec) {
-		expr->exec = NULL;
-	} else if (eval_fn == eval_fprintf) {
-		expr->printf = NULL;
-	} else if (eval_fn == eval_regex) {
-		expr->regex = NULL;
-	}
-
 	return expr;
 }
 
@@ -135,7 +112,7 @@ static struct bfs_expr *new_unary_expr(bfs_eval_fn *eval_fn, struct bfs_expr *rh
 
 	expr->lhs = NULL;
 	expr->rhs = rhs;
-	assert(bfs_expr_is_parent(expr));
+	bfs_assert(bfs_expr_is_parent(expr));
 
 	expr->persistent_fds = rhs->persistent_fds;
 	expr->ephemeral_fds = rhs->ephemeral_fds;
@@ -155,7 +132,7 @@ static struct bfs_expr *new_binary_expr(bfs_eval_fn *eval_fn, struct bfs_expr *l
 
 	expr->lhs = lhs;
 	expr->rhs = rhs;
-	assert(bfs_expr_is_parent(expr));
+	bfs_assert(bfs_expr_is_parent(expr));
 
 	expr->persistent_fds = lhs->persistent_fds + rhs->persistent_fds;
 	if (lhs->ephemeral_fds > rhs->ephemeral_fds) {
@@ -264,7 +241,7 @@ static void init_highlight(const struct bfs_ctx *ctx, bool *args) {
 static void highlight_args(const struct bfs_ctx *ctx, char **argv, size_t argc, bool *args) {
 	size_t i = argv - ctx->argv;
 	for (size_t j = 0; j < argc; ++j) {
-		assert(i + j < ctx->argc);
+		bfs_assert(i + j < ctx->argc);
 		args[i + j] = true;
 	}
 }
@@ -639,7 +616,7 @@ static const char *parse_int(const struct parser_state *state, char **arg, const
 		break;
 
 	default:
-		assert(!"Invalid int size");
+		bfs_bug("Invalid int size");
 		goto bad;
 	}
 
@@ -647,19 +624,19 @@ static const char *parse_int(const struct parser_state *state, char **arg, const
 
 bad:
 	if (!(flags & IF_QUIET)) {
-		parse_argv_error(state, arg, 1, "${bld}%s${rs} is not a valid integer.\n", str);
+		parse_argv_error(state, arg, 1, "${bld}%pq${rs} is not a valid integer.\n", str);
 	}
 	return NULL;
 
 negative:
 	if (!(flags & IF_QUIET)) {
-		parse_argv_error(state, arg, 1, "Negative integer ${bld}%s${rs} is not allowed here.\n", str);
+		parse_argv_error(state, arg, 1, "Negative integer ${bld}%pq${rs} is not allowed here.\n", str);
 	}
 	return NULL;
 
 range:
 	if (!(flags & IF_QUIET)) {
-		parse_argv_error(state, arg, 1, "${bld}%s${rs} is too large an integer.\n", str);
+		parse_argv_error(state, arg, 1, "${bld}%pq${rs} is too large an integer.\n", str);
 	}
 	return NULL;
 }
@@ -1082,16 +1059,16 @@ static struct bfs_expr *parse_time(struct parser_state *state, int field, int ar
 		switch (*tail) {
 		case 'w':
 			time *= 7;
-			BFS_FALLTHROUGH;
+			fallthru;
 		case 'd':
 			time *= 24;
-			BFS_FALLTHROUGH;
+			fallthru;
 		case 'h':
 			time *= 60;
-			BFS_FALLTHROUGH;
+			fallthru;
 		case 'm':
 			time *= 60;
-			BFS_FALLTHROUGH;
+			fallthru;
 		case 's':
 			break;
 		default:
@@ -1149,7 +1126,7 @@ static struct bfs_expr *parse_color(struct parser_state *state, int color, int a
 
 	if (color) {
 		if (!colors) {
-			parse_expr_error(state, expr, "%s.\n", strerror(ctx->colors_error));
+			parse_expr_error(state, expr, "Error parsing $$LS_COLORS: %s.\n", strerror(ctx->colors_error));
 			bfs_expr_free(expr);
 			return NULL;
 		}
@@ -1637,6 +1614,31 @@ static struct bfs_expr *parse_inum(struct parser_state *state, int arg1, int arg
 	return parse_test_icmp(state, eval_inum);
 }
 
+/**
+ * Parse -j<n>.
+ */
+static struct bfs_expr *parse_jobs(struct parser_state *state, int arg1, int arg2) {
+	struct bfs_expr *expr = parse_nullary_flag(state);
+	if (!expr) {
+		return NULL;
+	}
+
+	unsigned int n;
+	if (!parse_int(state, expr->argv, expr->argv[0] + 2, &n, IF_INT | IF_UNSIGNED)) {
+		bfs_expr_free(expr);
+		return NULL;
+	}
+
+	if (n == 0) {
+		parse_expr_error(state, expr, "${bld}0${rs} is not enough threads.\n");
+		bfs_expr_free(expr);
+		return NULL;
+	}
+
+	state->ctx->threads = n;
+	return expr;
+}
+
 /**
  * Parse -links N.
  */
@@ -1682,16 +1684,18 @@ static struct bfs_expr *parse_fnmatch(const struct parser_state *state, struct b
 		return NULL;
 	}
 
+	expr->pattern = expr->argv[1];
+
 	if (casefold) {
 #ifdef FNM_CASEFOLD
-		expr->num = FNM_CASEFOLD;
+		expr->fnm_flags = FNM_CASEFOLD;
 #else
 		parse_expr_error(state, expr, "Missing platform support.\n");
 		bfs_expr_free(expr);
 		return NULL;
 #endif
 	} else {
-		expr->num = 0;
+		expr->fnm_flags = 0;
 	}
 
 	// POSIX says, about fnmatch():
@@ -1700,10 +1704,9 @@ static struct bfs_expr *parse_fnmatch(const struct parser_state *state, struct b
 	//     return a non-zero value (indicating either no match or an error).
 	//
 	// But not all implementations obey this, so check for it ourselves.
-	const char *pattern = expr->argv[1];
-	size_t i, len = strlen(pattern);
+	size_t i, len = strlen(expr->pattern);
 	for (i = 0; i < len; ++i) {
-		if (pattern[len - i - 1] != '\\') {
+		if (expr->pattern[len - i - 1] != '\\') {
 			break;
 		}
 	}
@@ -1713,6 +1716,12 @@ static struct bfs_expr *parse_fnmatch(const struct parser_state *state, struct b
 		return expr;
 	}
 
+	// strcmp() can be much faster than fnmatch() since it doesn't have to
+	// parse the pattern, so special-case patterns with no wildcards.
+	//
+	//     https://pubs.opengroup.org/onlinepubs/9699919799/utilities/V3_chap02.html#tag_18_13_01
+	expr->literal = strcspn(expr->pattern, "?*\\[") == len;
+
 	return expr;
 }
 
@@ -1807,7 +1816,7 @@ static int parse_reftime(const struct parser_state *state, struct bfs_expr *expr
 static struct bfs_expr *parse_newerxy(struct parser_state *state, int arg1, int arg2) {
 	const char *arg = state->argv[0];
 	if (strlen(arg) != 8) {
-		parse_error(state, "Expected ${blu}-newer${bld}XY${rs}; found ${blu}-newer${bld}%s${rs}.\n", arg + 6);
+		parse_error(state, "Expected ${blu}-newer${bld}XY${rs}; found ${blu}-newer${bld}%pq${rs}.\n", arg + 6);
 		return NULL;
 	}
 
@@ -1962,11 +1971,11 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 		MODE_PLUS,
 		MODE_MINUS,
 		MODE_EQUALS,
-	} BFS_UNINIT(op, MODE_EQUALS);
+	} op = uninit(op, MODE_EQUALS);
 
-	mode_t BFS_UNINIT(who, 0);
-	mode_t BFS_UNINIT(file_change, 0);
-	mode_t BFS_UNINIT(dir_change, 0);
+	mode_t who = uninit(who, 0);
+	mode_t file_change = uninit(file_change, 0);
+	mode_t dir_change = uninit(dir_change, 0);
 
 	const char *i = mode;
 	while (true) {
@@ -1974,7 +1983,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 		case MODE_CLAUSE:
 			who = 0;
 			mstate = MODE_WHO;
-			BFS_FALLTHROUGH;
+			fallthru;
 
 		case MODE_WHO:
 			switch (*i) {
@@ -2001,7 +2010,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 			case MODE_EQUALS:
 				expr->file_mode &= ~who;
 				expr->dir_mode &= ~who;
-				BFS_FALLTHROUGH;
+				fallthru;
 			case MODE_PLUS:
 				expr->file_mode |= file_change;
 				expr->dir_mode |= dir_change;
@@ -2011,7 +2020,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				expr->dir_mode &= ~dir_change;
 				break;
 			}
-			BFS_FALLTHROUGH;
+			fallthru;
 
 		case MODE_ACTION:
 			if (who == 0) {
@@ -2093,7 +2102,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				break;
 			case 'x':
 				file_change |= who & 0111;
-				BFS_FALLTHROUGH;
+				fallthru;
 			case 'X':
 				dir_change |= who & 0111;
 				break;
@@ -2156,7 +2165,7 @@ static struct bfs_expr *parse_perm(struct parser_state *state, int field, int ar
 			++mode;
 			break;
 		}
-		BFS_FALLTHROUGH;
+		fallthru;
 	default:
 		expr->mode_cmp = BFS_MODE_EQUAL;
 		break;
@@ -2302,7 +2311,7 @@ static struct bfs_expr *parse_regextype(struct parser_state *state, int arg1, in
 		state->regex_type = BFS_REGEX_POSIX_BASIC;
 	} else if (strcmp(type, "posix-extended") == 0) {
 		state->regex_type = BFS_REGEX_POSIX_EXTENDED;
-#if BFS_WITH_ONIGURUMA
+#if BFS_USE_ONIGURUMA
 	} else if (strcmp(type, "emacs") == 0) {
 		state->regex_type = BFS_REGEX_EMACS;
 	} else if (strcmp(type, "grep") == 0) {
@@ -2324,7 +2333,7 @@ list_types:
 	cfprintf(cfile, "  ${bld}posix-basic${rs}:    POSIX basic regular expressions (BRE)\n");
 	cfprintf(cfile, "  ${bld}posix-extended${rs}: POSIX extended regular expressions (ERE)\n");
 	cfprintf(cfile, "  ${bld}ed${rs}:             Like ${grn}ed${rs} (same as ${bld}posix-basic${rs})\n");
-#if BFS_WITH_ONIGURUMA
+#if BFS_USE_ONIGURUMA
 	cfprintf(cfile, "  ${bld}emacs${rs}:          Like ${grn}emacs${rs}\n");
 	cfprintf(cfile, "  ${bld}grep${rs}:           Like ${grn}grep${rs}\n");
 #endif
@@ -2479,7 +2488,7 @@ static struct bfs_expr *parse_size(struct parser_state *state, int arg1, int arg
 	return expr;
 
 bad_unit:
-	parse_expr_error(state, expr, "Expected a size unit (one of ${bld}cwbkMGTP${rs}); found ${err}%s${rs}.\n", unit);
+	parse_expr_error(state, expr, "Expected a size unit (one of ${bld}cwbkMGTP${rs}); found ${err}%pq${rs}.\n", unit);
 fail:
 	bfs_expr_free(expr);
 	return NULL;
@@ -2754,7 +2763,9 @@ static struct bfs_expr *parse_help(struct parser_state *state, int arg1, int arg
 	cfprintf(cout, "      Enable optimization level ${bld}N${rs} (default: ${bld}3${rs})\n");
 	cfprintf(cout, "  ${cyn}-S${rs} ${bld}bfs${rs}|${bld}dfs${rs}|${bld}ids${rs}|${bld}eds${rs}\n");
 	cfprintf(cout, "      Use ${bld}b${rs}readth-${bld}f${rs}irst/${bld}d${rs}epth-${bld}f${rs}irst/${bld}i${rs}terative/${bld}e${rs}xponential ${bld}d${rs}eepening ${bld}s${rs}earch\n");
-	cfprintf(cout, "      (default: ${cyn}-S${rs} ${bld}bfs${rs})\n\n");
+	cfprintf(cout, "      (default: ${cyn}-S${rs} ${bld}bfs${rs})\n");
+	cfprintf(cout, "  ${cyn}-j${bld}N${rs}\n");
+	cfprintf(cout, "      Search with ${bld}N${rs} threads in parallel (default: number of CPUs, up to ${bld}8${rs})\n\n");
 
 	cfprintf(cout, "${bld}Operators:${rs}\n\n");
 
@@ -3061,6 +3072,7 @@ static const struct table_entry parse_table[] = {
 	{"-ipath", T_TEST, parse_path, true},
 	{"-iregex", T_TEST, parse_regex, BFS_REGEX_ICASE},
 	{"-iwholename", T_TEST, parse_path, true},
+	{"-j", T_FLAG, parse_jobs, 0, 0, true},
 	{"-links", T_TEST, parse_links},
 	{"-lname", T_TEST, parse_lname, false},
 	{"-ls", T_ACTION, parse_ls},
@@ -3493,7 +3505,7 @@ static const char *bftw_strategy_name(enum bftw_strategy strategy) {
 		return "eds";
 	}
 
-	assert(!"Invalid strategy");
+	bfs_bug("Invalid strategy");
 	return "???";
 }
 
@@ -3553,6 +3565,10 @@ void bfs_ctx_dump(const struct bfs_ctx *ctx, enum debug_flags flag) {
 		cfprintf(cerr, " ${cyn}-O${bld}%d${rs}", ctx->optlevel);
 	}
 
+	if (ctx->threads > 0) {
+		cfprintf(cerr, " ${cyn}-j${bld}%d${rs}", ctx->threads);
+	}
+
 	cfprintf(cerr, " ${cyn}-S${rs} ${bld}%s${rs}", bftw_strategy_name(ctx->strategy));
 
 	enum debug_flags debug = ctx->debug;
@@ -3577,7 +3593,7 @@ void bfs_ctx_dump(const struct bfs_ctx *ctx, enum debug_flags flag) {
 		if (c == '-' || c == '(' || c == ')' || c == '!' || c == ',') {
 			cfprintf(cerr, " ${cyn}-f${rs}");
 		}
-		cfprintf(cerr, " ${mag}%s${rs}", path);
+		cfprintf(cerr, " ${mag}%pq${rs}", path);
 	}
 
 	if (ctx->cout->colors) {
@@ -3643,14 +3659,11 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 	}
 
 	ctx->argc = argc;
-	ctx->argv = malloc((argc + 1)*sizeof(*ctx->argv));
+	ctx->argv = xmemdup(argv, sizeof_array(char *, argc + 1));
 	if (!ctx->argv) {
-		perror("malloc()");
+		perror("xmemdup()");
 		goto fail;
 	}
-	for (int i = 0; i <= argc; ++i) {
-		ctx->argv[i] = argv[i];
-	}
 
 	enum use_color use_color = COLOR_AUTO;
 	if (getenv("NO_COLOR")) {
@@ -3713,12 +3726,6 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 		.now = ctx->now,
 	};
 
-	const char *cmd = state.command + xbaseoff(state.command);
-	if (strcmp(cmd, "find") == 0) {
-		// Operate depth-first when invoked as "find"
-		ctx->strategy = BFTW_DFS;
-	}
-
 	ctx->exclude = bfs_expr_new(eval_false, 1, &fake_false_arg);
 	if (!ctx->exclude) {
 		goto fail;
@@ -3733,6 +3740,10 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 		}
 	}
 
+	if (state.use_color == COLOR_AUTO && !ctx->colors) {
+		bfs_warning(ctx, "Error parsing $$LS_COLORS: %s.\n\n", strerror(ctx->colors_error));
+	}
+
 	if (bfs_optimize(ctx) != 0) {
 		goto fail;
 	}
